/*
 * Copyright (C) 2020 Tycho Softworks.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "memory.hpp"

using namespace server;

mempager::mempager(std::size_t size) noexcept :
page_size(aligned_page(size)), mem_align(aligned_page()), page_count(0) {
    if(page_size <= mem_align)
        mem_align = aligned_cache();

    current_ = nullptr;
}

mempager::mempager(mempager&& move) noexcept :
page_size(move.page_size), mem_align(move.mem_align), page_count(move.page_count), current_(move.current_) {
    move.current_ = nullptr;
    move.page_count = 0;
}

mempager::~mempager() {
    clear();
}

auto mempager::operator=(mempager&& move) noexcept -> mempager& {
    page_size = move.page_size;
    mem_align = move.mem_align;
    page_count = move.page_count;
    current_ = move.current_;
    move.current_ = nullptr;
    move.page_count = 0;
    return *this;
}

auto mempager::alloc(std::size_t size) -> void * {
    assert(size > 0);

    while(size % sizeof(void *))
        ++size;

    if(size > (page_size - sizeof(page_ptr)))
        return nullptr;

    if(!current_ || size > page_size - current_->used) {
        auto page = static_cast<page_ptr>(aligned_alloc(mem_align, page_size));
        if(!page)
            return nullptr;
        page->used = sizeof(page_t);
        page->next = current_;
        ++page_count;
        current_ = page;
    }

    uint8_t *mem = (reinterpret_cast<uint8_t *>(current_)) + current_->used;
    current_->used += static_cast<unsigned>(size);
    return mem;
}

auto mempager::dup(const char *str) -> char * {
    if(!str)
        return nullptr;
    auto size = mem_size(str, FILENAME_MAX) + 1;
    auto mem = static_cast<char *>(alloc(size));
    memcpy(mem, str, size); // FlawFinder: ignore
    return mem;
}

auto mempager::dup(const std::string& str) -> char * {
    auto size = str.length() + 1;
    auto mem = static_cast<char *>(alloc(size));
    memcpy(mem, str.c_str(), size); // FlawFinder: ignore
    return mem;
}

void mempager::clear() {
    page_ptr next{};
    while(current_) {
        next = current_->next;
        free(current_); // NOLINT
        current_ = next;
    }
    page_count = 0;
}

auto server::mem_index(const uint8_t *mem, std::size_t size) -> unsigned {
    assert(mem != nullptr);

    unsigned val = 0;

    while(size--)
        val = (val << 1) ^ (*(mem++) & 0x1f);

    return val;
}

auto server::mem_size(const char *cp, std::size_t max) -> std::size_t {
    std::size_t count = 0;
    while(cp && *cp && count++ <= max)
        ++cp;

    if(count > max)
        throw std::runtime_error("memory size too large");

    return count;
}

auto server::mem_append(char *dest, std::size_t max, const char *value) -> std::size_t {
    assert(dest != nullptr);
    assert(value != nullptr);
    assert(max > 0);

    while(max && *dest) {
        ++dest;
        --max;
    }
    if(!max)
        throw std::runtime_error("memory target broken");

    --max; // null byte at end...

    std::size_t pos = 0;
    while(*value && pos < max) {
        *(dest++) = *(value++);
        ++pos;
    }
    *dest = 0;
    return pos;
}

auto server::mem_view(char *target, std::size_t size, std::string_view s) -> bool {
    auto count = s.size();
    auto from = s.data();

    if(count >= size)
        return false;

    while(count--)
        *(target++) = *(from++);
    *target = 0;
    return true;
}

auto server::mem_copy(char *target, std::size_t size, const char *from) -> bool {
    auto count = strnlen(from, size);
    if(count == size)
        return false;

    while(count--)
        *(target++) = *(from++);
    *target = 0;
    return true;
}

auto server::mem_value(char *target, std::size_t size, unsigned value) -> bool {
    auto max = 1U;
    auto zero = false;

    while(--size)
        max *= 10;

    if(value >= max)
        return false;

    while(max) {
        auto dig = 0U;
        max /= 10;
        if(max) {
            dig = value / max;
            value %= max;
        }
        if(dig || zero) {
            *(target++) = static_cast<char>('0' + dig);
            zero = true;
        }
        if(max == 1)
            break;
    }
    *target = 0;
    return true;
}

#ifdef __clang__
auto mempager::aligned_alloc(std::size_t align, std::size_t size) -> void * {
    void *addr = nullptr;
    if(posix_memalign(&addr, align, size))
        return nullptr;
    return addr;
}
#else
auto mempager::aligned_alloc(std::size_t align, std::size_t size) -> void * {
    return ::aligned_alloc(align, size);
}
#endif
