/*
 * Copyright (C) 2020 Tycho Softworks.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "drivers.hpp"
#include "network.hpp"

#include <moderncli/strings.hpp>
#include <moderncli/socket.hpp>

using namespace server;

auto net::to_string(const std::string& address, int port) -> std::string {
    auto host(address);

    if(address.empty() && ipv6)
        host = "[::]";
    else if(address.empty())
        host = "0.0.0.0";
    else if(address.find_first_of(':') != std::string::npos && address[0] != '[')
        host = "[" + address + "]";

    if(!port)
        return host;

    return host + ":" + std::to_string(port);
}

auto net::to_address(const std::string& name) -> std::string {
    using namespace std::literals::string_literals;
    struct addrinfo hints{}, *info{nullptr};

    memset(&hints, 0, sizeof(hints));
    hints.ai_family = ipv6 ? AF_INET6 : AF_INET;
    hints.ai_flags |= AI_ADDRCONFIG;
    if(getaddrinfo(name.c_str(), nullptr, &hints, &info) != 0)
        return {};

    if(info && info->ai_addr) {
        auto host = inet_host(info->ai_addr);
        if(host.empty())
            return host;
        if(ipv6)
            return "[" + host + "]";
        return host;
    }
    return "";
}

auto net::uri_schema(const std::string& id) -> std::string {
    if(begins_with(id, "sip:"))
        return "sip:";
    if(begins_with(id, "sips:"))
        return "sips:";
    if(begins_with(id, "h323:"))
        return "h323:";
    return {};
}

auto net::uri_userid(const std::string& id) -> std::string {
    auto user(id);
    if(begins_with(id, "sip:"))
        user = id.substr(4);
    else if(begins_with(id, "sips:") || begins_with(id, "h323:"))
        user = id.substr(5);

    auto pos = user.find_last_of('@');
    if(pos != std::string::npos)
        return user.substr(0, pos);
    return user;
}

auto net::uri_server(const std::string& id) -> std::string {
    auto server(id);
    if(begins_with(id, "sip:"))
        server = id.substr(4);
    else if(begins_with(id, "sips:") || begins_with(id, "h323:"))
        server = id.substr(5);

    auto pos = server.find_last_of('@');
    if(pos != std::string::npos)
        return server.substr(pos + 1);
    return server;
}

auto net::uri_host(const std::string& id) -> std::string {
    auto server(id);
    auto pos = id.find_last_of('@');
    if(pos != std::string::npos)
        server = id.substr(pos + 1);
    else {
        if(begins_with(id, "sip:"))
            server = id.substr(4);
        else if(begins_with(id, "sips:") || begins_with(id, "h323:"))
            server = id.substr(5);
    }

    if(begins_with(id, "h323:"))
        return server;

    auto inet6 = server.find_last_of(']');
    pos = server.find_last_of(':');

    if((pos != std::string::npos) && (inet6 == std::string::npos || inet6 < pos))
        return server.substr(0, pos);

    return server;
}

auto net::uri_address(const std::string& id) -> std::string {
    auto host = uri_host(id);
    auto addr = to_address(host);

    if(host.empty() || addr.empty())
        return host;

    return addr;
}

auto net::uri_port(const std::string& id) -> uint16_t {
    auto server(id);
    auto port = uint16_t(5060);

    if(begins_with(id, "sip:"))
        server = id.substr(4);
    else if(begins_with(id, "h323:"))
        return 1720;
    else if(begins_with(id, "sips:")) {
        server = id.substr(5);
        ++port;
    }

    auto inet6 = server.find_last_of(']');
    auto pos = server.find_last_of(':');

    if(pos == std::string::npos)
        return port;

    if(inet6 != std::string::npos && inet6 > pos)
        return port;

    return std::stoi(server.substr(pos + 1));
}

auto net::set_user(const std::string& uri, const std::string& user) -> std::string {
    auto schema = uri_schema(uri);
    auto server = uri_server(uri);

    auto inet6 = server.find_last_of(']');
    auto pos = server.find_last_of(':');

    if(schema == "h323:")   // we never use ports in h323
        inet6 = pos;

    if((pos != std::string::npos) && (inet6 == std::string::npos || inet6 < pos)) {
        auto port = std::stoi(server.substr(pos + 1));
        return make_uri(schema, user, server.substr(0, pos), port);               }
    return make_uri(schema, user, server);
}

auto net::make_uri(const std::string& schema, const std::string& host, uint16_t port) -> std::string {
    if(schema == "h323:" || !port)
        return schema + host;

    return schema + host + ":" + std::to_string(port);
}

auto net::make_uri(const std::string& schema, const std::string& user, const std::string& host, uint16_t port) -> std::string {
    if(schema == "h323:" || !port)
        return schema + user + "@" + host;

    return schema + user + "@" + host + ":" + std::to_string(port);
}

