/*
 * Copyright (C) 2021 Tycho Softworks.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "server.hpp"       // IWYU pragma: keep

#include "ipc.hpp"

#include <moderncli/args.hpp>

#include <mqueue.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>

namespace {
args::option helpflag('h',  "--help",   "display this list");
args::option quiet('q',     "--quiet",  "suppress most output");
args::option althelp('?',   nullptr, nullptr);

class ipc_error final : public std::exception {
public:
    explicit ipc_error(const char *text, int code) noexcept : msg(text), err(code) {}

    auto what() const noexcept -> const char * override {
        return msg;
    }

    auto exit() const noexcept {
        return err;
    }

private:
    const char *msg;
    int err;
};

enum {STATUS, LEVEL, SESSIONS, RELOAD, INVALID};
auto command = INVALID;
auto exit_code = 0;
} // end namespace

auto main(int argc, const char **argv) -> int { // NOLINT
    try {
        auto count = args::parse(argc, argv);
        if(count == 1 && args::list()[0] == "status")
            command = STATUS;

        if(count == 1 && args::list()[0] == "reload")
            command = RELOAD;

        if(count == 1 && args::list()[0] == "list")
            command = SESSIONS;

        if(count == 2 && args::list()[0] == "level")
            command = LEVEL;

        if(is(helpflag) || is(althelp) || !count || command == INVALID) {
            args::help({"bordeauxctl [options] command..."}, {
                "bordeauxctl " PROJECT_VERSION,
                "David Sugar <tychosoft@gmail.com>",
                "Send command to server ipc"
            });
            return 0;
        }

        ipc_event event{.type = IPC_STATUS, .ver = IPC_VERSION};
        if(command == LEVEL) {
            event.type = IPC_LEVEL;
            event.body.level = std::stoi(std::string{args::list()[1]});
        }
        else if(command == RELOAD) {
            event.type = IPC_RELOAD;
        }

        auto mq = mq_open(BORDEAUX_IPCPATH, O_WRONLY, 0660, nullptr);
        if(mq == MQD_INVALID)
            throw ipc_error("mqueue failed", 4);
        if(mq_send(mq, reinterpret_cast<const char *>(&event), sizeof(event), 0) < 0)
            throw ipc_error("send failed", 4);

        auto shm = shm_open(BORDEAUX_IPCPATH, O_RDONLY, 0660);
        if(shm < 0)
            throw ipc_error("mapping failed", 4);

        struct stat ino{};
        if(fstat(shm, &ino) < 0)
            throw ipc_error("stat failed", 4);
        auto max = (ino.st_size - sizeof(ipc_system)) / sizeof(ipc_session);
        auto mem = mmap(nullptr, sizeof(ipc_session) * max + sizeof(ipc_system), PROT_READ, MAP_SHARED, shm, 0);
        if(mem == MAP_FAILED)   // NOLINT
            throw ipc_error("mapping failed", 4);

        auto sys = static_cast<ipc_system*>(mem);
        auto map = reinterpret_cast<ipc_session*>(&sys[1]);
        auto used = sys->used.load();
        if(used > sys->limit)
            used = sys->limit;

        if(!sys->active && command == STATUS)
            throw ipc_error("offline", 10);

        time_t now{};
        time(&now);

        if(command == SESSIONS) {
            for(auto idx = 0U; idx < max; ++idx) {
                const auto& entry = map[idx];
                if(!entry.created)
                    continue;

                char state = '?';
                char type = '?';
                switch(entry.state) {
                case ipc_session::ISC_RINGING:
                    state = 'R';
                    break;
                case ipc_session::ISC_CLOSING:
                    state = 'X';
                    break;
                case ipc_session::ISC_INVITING:
                    state = 'I';
                    break;
                case ipc_session::ISC_CONNECTED:
                    state = 'C';
                    break;
                default:
                    break;
                }

                switch(entry.type) {
                case ipc_session::ISC_INBOUND:
                    type = 'I';
                    break;
                case ipc_session::ISC_OUTBOUND:
                    type = 'O';
                    break;
                default:
                    break;
                }

                print("{} {}{} {} {} {}s\n", entry.sid, type, state, entry.caller, entry.dialed, now - entry.created);
            }
        }

        now -= sys->started;
        if(!is(quiet))
            print("used={}/{} uptime={}m log={}\n", used, sys->limit, now / 60, sys->level);

    }
    catch(const bad_arg& e) {
        die(-1, "Usage: bordeauxctl [options] command...\nerror: {}\n", e.what());
    }
    catch(const ipc_error& e) {
        exit_code = e.exit();
        if(!is(quiet))
            die(exit_code, "error: {}\n", e.what());
    }
    catch(const std::exception& e) {
        if(is(quiet))
            exit_code = -1;
        else
            die(-1, "error: {}\n", e.what());
    }
    catch(...) {
        if(is(quiet))
            exit_code = -1;
        else
            die(-1, "error: unknown error\n");
    }
    return exit_code;
}

